package com.example.aspdemo.mediastreamplayer.video;

import android.media.MediaCodec;
import android.media.MediaFormat;
import android.os.Build;
import android.util.Log;
import android.view.Surface;

import com.aliyun.wuying.aspsdk.aspengine.VCodecType;
import com.aliyun.wuying.aspsdk.aspengine.VFrame;
import com.aliyun.wuying.aspsdk.aspengine.VProfile;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Queue;

class MediaCodecVideoDecoder {
    private static final String TAG = "MediaCodecVideoDecoder";

    private final long DEQUEUE_BUFFER_TIMEOUT = 1000000000;
    private MediaCodec mDecoder = null;
    private MediaFormat mFormat = null;
    private Thread mDecoderOutputHandler = null;
    private boolean mOutputThreadRunning = false;
    private boolean mConfigured = false;
    private long mAvgDecodeTime = 0;
    private long mTotalDecodeTime = 0;
    private long mDecodedTime = 0;
    private long mMinDecodeTime = 0;
    private long mMaxDecodeTime = 0;

    private void handleOutputFrame() {
        if (mDecoder != null) {
            MediaCodec.BufferInfo info = new MediaCodec.BufferInfo();
            int outIndex = mDecoder.dequeueOutputBuffer(info, -1);
            if (outIndex >= 0) {
                long startTime = 0;
                synchronized (this) {
                    startTime = mDecodeTimeList.get(0);
                    mDecodeTimeList.remove(0);
                }
                mDecodedTime++;
                long decodeTimeConsumed = (System.nanoTime() - startTime) / 1000 / 1000;
                mTotalDecodeTime += decodeTimeConsumed;
                mAvgDecodeTime = mTotalDecodeTime / mDecodedTime;
                mMinDecodeTime = mMinDecodeTime == 0 ? decodeTimeConsumed : Math.min(mMinDecodeTime, decodeTimeConsumed);
                mMaxDecodeTime = Math.max(mMaxDecodeTime, decodeTimeConsumed);
                Log.i(TAG, "Decode time " + decodeTimeConsumed
                        + ", " + info.presentationTimeUs
                        + ", avg " + mAvgDecodeTime
                        + ", min " + mMinDecodeTime
                        + ", max " + mMaxDecodeTime
                        + ", frame " + mDecodedTime
                        + ", start time " + startTime);
                if ((info.flags & MediaCodec.BUFFER_FLAG_END_OF_STREAM) != 0) {
                    Log.i(TAG, "output EOS");
                }
                mDecoder.releaseOutputBuffer(outIndex, true);
            }
            Log.i(TAG, "dequeueOutputBuffer index " + outIndex);
        }
    }

    class DecoderOutputHandler implements Runnable {
        @Override
        public void run() {
            while (mOutputThreadRunning) {
                handleOutputFrame();
            }
        }
    }

    public boolean setVideoProfile(VProfile profile) {
        if (profile.codecType == VCodecType.H264) {
            mFormat = new MediaFormat();
            mFormat.setString(MediaFormat.KEY_MIME, "video/avc");
            mFormat.setInteger(MediaFormat.KEY_WIDTH, profile.width);
            mFormat.setInteger(MediaFormat.KEY_HEIGHT, profile.height);
            mFormat.setInteger(MediaFormat.KEY_PRIORITY, 0);
            mFormat.setInteger(MediaFormat.KEY_LOW_LATENCY, 1);

            try {
                mDecoder = MediaCodec.createDecoderByType("video/avc");
            } catch (Exception ex) {
                Log.e(TAG, "failed to create video decoder");
                return false;
            }
        }

        return true;
    }

    public boolean setSurface(Surface surface) {
        if (!mConfigured && surface != null) {
            mDecoder.configure(mFormat, surface, null, 0);
            mConfigured = true;
        } else if (mConfigured) {
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                mDecoder.setOutputSurface(surface);
            } else {
                // Otherwise, need to rebuild the mediacodec
            }
        }

        return true;
    }

    public boolean start() {
        if (mDecoder != null && !mOutputThreadRunning) {
            mDecoder.start();
            mOutputThreadRunning = true;
            mDecoderOutputHandler = new Thread(new DecoderOutputHandler());
            mDecoderOutputHandler.start();
        }

        return true;
    }

    public boolean stop() {
        if (mDecoder != null) {
            mOutputThreadRunning = false;
            try {
                mDecoderOutputHandler.join();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            mDecoder.stop();
            mDecoder = null;
            mConfigured = false;
        }

        return true;
    }

    private ArrayList<Long> mDecodeTimeList = new ArrayList<>();
    private int mIndex = 0;

    public boolean handleInputFrame(VFrame frame) {
        if (mDecoder != null) {
            int tryTimes = 0;
            do {
                Log.e(TAG, "Start dequeue input buffer");
                int index = mDecoder.dequeueInputBuffer(-1);
                if (index >= 0) {
                    ByteBuffer buffer = mDecoder.getInputBuffer(index);
                    if (buffer == null) {
                        Log.e(TAG, "No available codec buffer, index " + index);
                        return false;
                    } else {
                        buffer.put(frame.frame);
                        synchronized (this) {
                            long startTime = System.nanoTime();
                            mDecodeTimeList.add(startTime);
                            Log.d(TAG, "buffer queued," + index + ", size " + frame.size + ", index " + mIndex
                                    + ", start time " + startTime);
                        }
                        frame.mmTime = 0;
                        mDecoder.queueInputBuffer(index, 0, frame.size, 0, 0);
                        mIndex++;

                        Log.e(TAG, "Handle output frame done");
                        break;
                    }
                } else {
                    try {
                        Thread.sleep(5);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            } while (tryTimes++ < 5);
        }

        return true;
    }
}
